
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <termios.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include "localserver.h"


//#define DEBUG_BLUESERVER

#ifdef DEBUG_BLUESERVER
  #define LS_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LS_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LS_DEBUG(msg...)     do { } while (0)
  #define LS_DEBUGT(t, msg...) do { } while (0)
#endif


// major class of device codes
// taken from: https://www.bluetooth.org/foundry/assignnumb/document/baseband
#define MAJ_COD_MISC                  0
#define MAJ_COD_COMPUTER              1
#define MAJ_COD_PHONE                 2
#define MAJ_COD_LAN                   3
#define MAJ_COD_AUDIO_VIDEO           4
#define MAJ_COD_COMPUTER_PERIPHERAL   5
#define MAJ_COD_IMAGING               6
#define MAJ_COD_UNCATEGORIZED         31

// minor class of device types (where MajorDeviceType is Computer or PDA)
#define MAJ_COMP_MIN_HANDHELD         0x04  // Handheld PC/PDA (clam shell)
#define MAJ_COMP_MIN_PALMSIZED        0x05  // Palm sized PC/PDA

// minor class of device types (where MajorDeviceType is Audio or Video device)
#define MAJ_AV_MIN_HEADSETDEVICE      0x01  // Wearable Headset Device
#define MAJ_AV_MIN_LOUDSPEAKER        0x05  // Loudspeaker
#define MAJ_AV_MIN_HEADPHONE          0x06  // A2DP headphones
#define MAJ_AV_MIN_PORTABLE           0x07  // Portable audio
#define MAJ_AV_MIN_CARAUDIO           0x08  // Car audio
#define MAJ_AV_MIN_HIFIDEVICE         0x0A  // HiFi device
#define MAJ_AV_MIN_VIDEO_AND_SP       0x0F  // Video Display and Loudspeaker


#define DEBUG_CLASS_OF_DEVICE_FILTERING

char outbuf[1024];

int scan_profiles_fd = -1;
typedef int (*handler_t)(bdaddr_t *bdaddr, struct search_context *arg);
static bdaddr_t interface_bdaddr;
int estr2ba(char *str, bdaddr_t *ba)
{
  // Only trap "local", "any" is already dealt with 
  if(!strcmp(str, "local")) 
  {
    bacpy(ba, BDADDR_LOCAL);
    return 0;
  }
  return str2ba(str, ba);
}

// Pass args to the inquiry/search handler 
struct search_context 
{
  char      *svc;    // Service 
  uuid_t    group;   // Browse group 
  int       tree;    // Display full attribute tree 
  uint32_t  handle;  // Service record handle 
};



static char UUID_str[MAX_LEN_UUID_STR];

static void writestr(int fd, const char *str)
{
  size_t len = strlen(str) + 1; // add NUL terminator
  ssize_t nw = write(fd, str, len);
  if (nw < 0)
  {
    printf("blueserver: unable to write to socket: %s\n", strerror(errno));
  }
  else if (static_cast<size_t>(nw) != len)
  {
    printf("blueserver: only wrote %zd of %zu bytes to socket\n", nw, len);
  }
}


static void scan_profiles_print_service_class(void *value, void *userData)
{
  char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
  uuid_t *uuid = (uuid_t *)value;

  sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
  sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
  char dummy[128];
  sprintf(dummy,"0x%s|%s|", UUID_str, ServiceClassUUID_str);
  strcat(outbuf,dummy);
}


static void scan_profiles_print_service_desc(void *value, void *user)
{
  char str[MAX_LEN_PROTOCOL_UUID_STR];
  sdp_data_t *p = (sdp_data_t *)value, *s;
  int i = 0, proto = 0;
  
  for (; p; p = p->next, i++) 
  {
    switch (p->dtd) 
    {
    case SDP_UUID16:
    case SDP_UUID32:
    case SDP_UUID128:
      sdp_uuid2strn(&p->val.uuid, UUID_str, MAX_LEN_UUID_STR);
      sdp_proto_uuid2strn(&p->val.uuid, str, sizeof(str));
      proto = sdp_uuid_to_proto(&p->val.uuid);
      break;
    case SDP_UINT8:
      if (proto == RFCOMM_UUID)
      {
        sprintf(outbuf,"Channel|%d|\n", p->val.uint8);
        writestr(scan_profiles_fd,outbuf);
      }
      break;
    default:
      break;
    }
  }
}


void scan_profiles_print_access_protos(void* vr, void* userData)
{
  sdp_list_t *value = (sdp_list_t *)vr;
  sdp_list_t *protDescSeq = (sdp_list_t *)value;
  sdp_list_foreach(protDescSeq, scan_profiles_print_service_desc, 0);
}


//
// Parse a SDP record in user friendly form.
//
void scan_profiles_print_service_attr(sdp_record_t *rec)
{
  sdp_list_t *list = 0, *proto = 0;

  sdp_data_t *d = sdp_data_get(rec, SDP_ATTR_SVCNAME_PRIMARY);
  if (d) 
  {
    sprintf(outbuf,"Name|%s|\n", d->val.str);
    writestr(scan_profiles_fd,outbuf);
  }
  else
  {
    d = sdp_data_get(rec, SDP_ATTR_SVCDESC_PRIMARY);
    if (d) 
    {
      sprintf(outbuf,"Name|%s|\n", d->val.str);
      writestr(scan_profiles_fd,outbuf);
    }
    else
    {
      d = sdp_data_get(rec, SDP_ATTR_PROVNAME_PRIMARY);
      if (d) 
      {
        sprintf(outbuf,"Name|%s|\n", d->val.str);   
        writestr(scan_profiles_fd,outbuf);
      }
    }
  }

//  printf("Service RecHandle: 0x%x\n", rec->handle);

  if (sdp_get_service_classes(rec, &list) == 0) 
  {
//    printf("Service Class ID List:\n");
    sprintf(outbuf,"ServiceClass|");
    sdp_list_foreach(list, scan_profiles_print_service_class, 0);
    strcat(outbuf,"\n");
    writestr(scan_profiles_fd,outbuf);
    sdp_list_free(list, free);
  }
  if (sdp_get_access_protos(rec, &proto) == 0) 
  {
//    printf("Protocol Descriptor List:\n");
    sdp_list_foreach(proto, scan_profiles_print_access_protos, 0);
    sdp_list_free(proto, (sdp_free_func_t)sdp_data_free);
  }
}


//
// Perform an inquiry and search/browse all peer found.
//
static void scan_profiles_inquiry(handler_t handler, void *arg)
{
  inquiry_info ii[20];
  uint8_t count = 0;
  int i;

  LS_DEBUG("blueserver: Inquiring ...\n");
  if (sdp_general_inquiry(ii, 20, 8, &count) < 0) 
  {
    LS_DEBUG("blueserver: Inquiry failed\n");
    return;
  }
  for (i=0; i<count; i++)
    handler((bdaddr_t *)&ii[i].bdaddr, (struct search_context *)arg);
}

/*
 * Search for a specific SDP service
 */
int scan_profiles_do_search(bdaddr_t *bdaddr, struct search_context *context)
{
  sdp_list_t *attrid, *search, *seq, *next;
  uint32_t range = 0x0000ffff;
  char str[20];
  sdp_session_t *sess;

  if (!bdaddr) 
  {
    scan_profiles_inquiry(scan_profiles_do_search, context);
    return 0;
  }
  sess = sdp_connect(&interface_bdaddr, bdaddr, SDP_RETRY_IF_BUSY);
  ba2str(bdaddr, str);
  if (!sess) {
    LS_DEBUG("blueserver: Failed to connect to SDP server on %s: %s\n", str, strerror(errno));
    return -1;
  }

  attrid = sdp_list_append(0, &range);
  search = sdp_list_append(0, &context->group);
  if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
    LS_DEBUG("blueserver: Service Search failed: %s\n", strerror(errno));
    sdp_close(sess);
    return -1;
  }
  sdp_list_free(attrid, 0);
  sdp_list_free(search, 0);

  for (; seq; seq = next) 
  {
    sdp_record_t *rec = (sdp_record_t *) seq->data;
    struct search_context sub_context;

    if (context->tree) 
    {
      /* Display full tree */
//      sdp_printf_service_attr(rec);
    } 
    else 
    {
      /* Display user friendly form */
      scan_profiles_print_service_attr(rec);
    }
//    printf("\n");
    
    if (sdp_get_group_id(rec, &sub_context.group) != -1) 
    {
      /* Set the subcontext for browsing the sub tree */
      memcpy(&sub_context, context, sizeof(struct search_context));
      /* Browse the next level down if not done */
      if (sub_context.group.value.uuid16 != context->group.value.uuid16)
        scan_profiles_do_search(bdaddr, &sub_context);
    }
    next = seq->next;
    free(seq);
    sdp_record_free(rec);
  }
  sdp_close(sess);
//  printf("END\n");
  return 0;
}

#ifdef SUPPORT_BRISTOL
/*
 * Browse the SP profiles
 */
int scan_profiles_cmd_sp_browse(char* address)
{
  bacpy(&interface_bdaddr, BDADDR_ANY);

  struct search_context context;
  int opt;

  /* Initialise context */
  memset(&context, '\0', sizeof(struct search_context));
  /* Some phones don't return any services if you ask them for the public
   * browse group.  It turns out that the Windows XP Bluetooth browser asks
   * for L2CAP_UUID, which seems to work for all phones. */
  sdp_uuid16_create(&(context.group), SERIAL_PORT_SVCLASS_ID);

  bdaddr_t bdaddr;
  estr2ba(address, &bdaddr);
  return scan_profiles_do_search(&bdaddr, &context);
}
#endif  // SUPPORT_BRISTOL

/*
 * Browse the full SDP database (i.e. list all services starting from the
 * root/top-level).
 */
void scan_profiles_cmd_browse(char* address)
{
  bacpy(&interface_bdaddr, BDADDR_ANY);

  struct search_context context;
  int opt;
  bdaddr_t bdaddr;

  estr2ba(address, &bdaddr);

  /* Initialise context */
  memset(&context, '\0', sizeof(struct search_context));
#if 0
  sdp_uuid16_create(&(context.group), PUBLIC_BROWSE_GROUP);
  if (scan_profiles_do_search(&bdaddr, &context) != -1)
	return;
  LS_DEBUG("Scanning PUBLIC_BROWSE_GROUP failed!\n");
#else
  sdp_uuid16_create(&(context.group), L2CAP_UUID);
  if (scan_profiles_do_search(&bdaddr, &context) != -1)
	return;
  LS_DEBUG("Scanning L2CAP_UUID failed!\n");
#endif
  /* Scan for specific profiles if general scan fails */
  sdp_uuid16_create(&(context.group), HANDSFREE_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context)  == -1)
    LS_DEBUG("Scanning HANDSFREE_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), DIALUP_NET_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning DIALUP_NET_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), HEADSET_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning HEADSET_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), AUDIO_SINK_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning AUDIO_SINK_SVCLASS_ID failed!\n");

  return;
}

#define CSR_VARID_ENABLE_TX 0x4007
#define CSR_VARID_DISABLE_TX  0x4008

char *bccmd_err_names[] = {
  "ok",
  "no such variable id",
  "value too big",
  "no value",
  "bad request",
  "no access",
  "read only",
  "write only",
  "error",
  "permission denied",
};

int csr_write_varid_uint16(int dd, uint16_t seqnum, uint16_t varid, uint16_t value)
{
  unsigned char cmd[] = {
    0x02, 0x00,
    0x09, 0x00,
    seqnum & 0xff, seqnum >> 8,
    varid & 0xff, varid >> 8,
    0x00, 0x00,      // status
    0x00, 0x00,      // payload
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00
  };

  unsigned char cp[254], rp[254];
  struct hci_request rq;
  uint16_t status;

  memset(&cp, 0x00, sizeof(cp));
  cp[0] = 0xc2;
  memcpy(cp + 1, cmd, sizeof(cmd));

  memset(&rq, 0, sizeof(rq));
  rq.ogf    = OGF_VENDOR_CMD;
  rq.ocf    = 0x00;
  rq.event  = EVT_VENDOR;
  rq.cparam = cp;
  rq.clen   = sizeof(cmd) + 1;
  rq.rparam = rp;
  rq.rlen   = sizeof(rp);

  if(hci_send_req(dd, &rq, 2000) < 0) 
  {
    LS_DEBUG("blueserver: hci_send_req %d %s\n", errno, strerror(errno));
    return -1;
  }

  if(rp[0] != 0xc2) 
  {
    LS_DEBUG("blueserver: hci_send_req ret %d %s\n", errno, strerror(errno));
    errno = EIO;
    return -1;
  }

  status = rp[9] + (rp[10] << 8);
  if(status != 0) 
  {
    LS_DEBUG("blueserver: Illegal request: %s\n", bccmd_err_names[status]);
    return -1;
  }

  return 0;
}

int csr_enable_tx(int dd, uint16_t seqnum)
{
  return csr_write_varid_uint16(dd, seqnum, CSR_VARID_ENABLE_TX, 0);
}

int csr_disable_tx(int dd, uint16_t seqnum)
{
  return csr_write_varid_uint16(dd, seqnum, CSR_VARID_DISABLE_TX, 0);
}

void EnableBluetooth(int aEnable)
{
  LS_DEBUG("blueserver: EnableBluetooth called with %d\n",aEnable);
  int dev_id = -1;
  int ret=0;
  static int seq=1;
  int dd;

  dev_id = hci_devid("hci0");
  if(dev_id < 0) 
  {
    LS_DEBUG("blueserver: Invalid device\n");
    return;
  }
  dd = hci_open_dev(dev_id);
  if(dd < 0) 
  {
    LS_DEBUG("blueserver: Error opening hci device\n");
    return;
  }

  if(aEnable) 
  {
    LS_DEBUG("blueserver: disabling flight mode - enabling TX\n");
    ret = csr_enable_tx(dd, seq++);
  }
  else
  {
    LS_DEBUG("blueserver: enabling flight mode - disabling TX\n");
    ret = csr_disable_tx(dd, seq++);
  }

  if(ret)
  {
    LS_DEBUG("blueserver: ret=%d\n", ret);
  }
  hci_close_dev(dd);
}


static void requesthandler(char* request,int fd)
{
  if (!strncmp(request,"DISCOVER",8))
  {
    int headset_discovery = (strlen(request)==11 && !strncmp(request,"DISCOVER HS",11));
    int a2dp_discovery = (strlen(request)==13 && !strncmp(request,"DISCOVER A2DP",13));
    LS_DEBUGT(headset_discovery, "blueserver: headset discovery\n");
    LS_DEBUGT(a2dp_discovery, "blueserver: stereo headphones discovery\n");

    inquiry_info *info = NULL;
    int num_rsp, length, flags;
    char addr[18];
    char name[248];
    int i, opt, dd;

    length  = 8;  /* ~10 seconds */
    num_rsp = 100;
    flags = IREQ_CACHE_FLUSH;
    int dev_id = -1;
    if (dev_id < 0) 
    {
      dev_id = hci_get_route(NULL);
      if (dev_id < 0) 
      {
        perror("Device is not available");
        writestr(fd, "END");
        return;
      }
    }

    LS_DEBUG("blueserver: Scanning ...\n");
    int check_count = 0;
REDO:
    num_rsp = hci_inquiry(dev_id, length, num_rsp, NULL, &info, flags);
    if (num_rsp == -1 && check_count < 30)  // 30 sec retry period
    {
      LS_DEBUG("blueserver: RETRYING  num_rsp=%d  check_count=%d\n", num_rsp, check_count);
      sleep(1);
      check_count++;
      goto REDO;
    }

    LS_DEBUG("blueserver: num_rsp = %d\n",num_rsp);

    if (num_rsp >= 0) 
    {
       dd = hci_open_dev(dev_id);
      if (dd >= 0) 
      {
        for (i = 0; i < num_rsp; i++) 
        {
          int majorClassOfDevice =  (((info+i)->dev_class[1])&0x1f);
          int minorClassOfDevice = ((((info+i)->dev_class[0])>>2)&0x1f);

          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          LS_DEBUG("blueserver: majCOD=%d: ", majorClassOfDevice);
          LS_DEBUG("minCOD=%d: ", minorClassOfDevice);
          switch (majorClassOfDevice)
          {
            case MAJ_COD_MISC:                LS_DEBUG("Miscellaneous\n"); break;
            case MAJ_COD_COMPUTER:            LS_DEBUG("Computer (desktop,notebook, PDA, organizers)\n"); break;
            case MAJ_COD_PHONE:               LS_DEBUG("Phone (cellular, cordless, payphone, modem)\n"); break;
            case MAJ_COD_LAN:                 LS_DEBUG("LAN/Network Access point\n"); break;
            case MAJ_COD_AUDIO_VIDEO:         LS_DEBUG("Audio/Video (headset,speaker,stereo, video display, vcr)\n"); break;
            case MAJ_COD_COMPUTER_PERIPHERAL: LS_DEBUG("Peripheral (mouse, joystick, keyboards)\n"); break;
            case MAJ_COD_IMAGING:             LS_DEBUG("Imaging (printing, scanner, camera, display)\n"); break;
            case MAJ_COD_UNCATEGORIZED:       LS_DEBUG("Uncategorized, specific device code not specified\n"); break;
            default:                          LS_DEBUG("Unknown class of device\n"); break;
          }
          #endif
          
          int skipType = 1;
          if (headset_discovery)
          {
            if (majorClassOfDevice == MAJ_COD_AUDIO_VIDEO && 
                minorClassOfDevice == MAJ_AV_MIN_HEADSETDEVICE)
              skipType = 0;
          }
          else if (a2dp_discovery)
          {
            if (majorClassOfDevice == MAJ_COD_AUDIO_VIDEO && 
               (minorClassOfDevice == MAJ_AV_MIN_HEADSETDEVICE ||
                minorClassOfDevice == MAJ_AV_MIN_HEADPHONE     ||
                minorClassOfDevice == MAJ_AV_MIN_LOUDSPEAKER   ||
                minorClassOfDevice == MAJ_AV_MIN_HEADPHONE     ||
                minorClassOfDevice == MAJ_AV_MIN_PORTABLE      ||
                minorClassOfDevice == MAJ_AV_MIN_CARAUDIO      ||
                minorClassOfDevice == MAJ_AV_MIN_HIFIDEVICE    ||
                minorClassOfDevice == MAJ_AV_MIN_VIDEO_AND_SP))
              skipType = 0;
          }
          else
          {
            if (majorClassOfDevice == MAJ_COD_PHONE)
              skipType = 0;
            if (majorClassOfDevice == MAJ_COD_COMPUTER && 
               (minorClassOfDevice == MAJ_COMP_MIN_HANDHELD || 
                minorClassOfDevice == MAJ_COMP_MIN_PALMSIZED)) 
              skipType = 0;
          }
          
          if (skipType)
          {  
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("blueserver: Skipping (not a phone/PDA, headset or HiFi device)\n");
            #endif

            continue;
          }

          // If the client sends *any* data, just stop scanning
          {
            int nread;
            ioctl(fd, FIONREAD, &nread);
            if (nread>0)
            {
              LS_DEBUG("blueserver: Stopping scanning for bluetooth devices\n");
              break;
            }
          }
          
          
          memset(name, 0, sizeof(name));
          if (hci_read_remote_name(dd, &(info+i)->bdaddr, sizeof(name), name, 100000) < 0)
          {
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("blueserver: Unable to read device name: hci_read_remote_name failed\n");
            #endif

            // Skip devices with no name
//            strcpy(name, "n/a");
            continue;
          }
          ba2str(&(info+i)->bdaddr, addr);
          char buf[256];
          snprintf(buf,256,"%s|%s", addr, name);
          writestr(fd, buf);

          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          LS_DEBUG("blueserver: name = %s\n", buf);
          #endif
        }
        close(dd);
        free(info);
      }
    }
    writestr(fd, "END");
  }
  else if (!strncmp(request,"PROFILES",8))
  {
    #define BT_CLIENT_PINFILE   "/etc/bluetooth/client_pin"

    // "PROFILE bdaddr" where bdaddr is macid of client
    // "PROFILE [HS|HF]|pin|bdaddr" where pin is the client pincode to use on bdaddr
    int profiles_a2dp = (strlen(request)>=14 && !strncmp(request,"PROFILES A2DP|",14));
    int profiles_headset = (strlen(request)>=12 && !strncmp(request,"PROFILES HS|",12));
    int profiles_protect = (strlen(request)>=12 && !strncmp(request,"PROFILES HF|",12));
#ifdef SUPPORT_BRISTOL
    int profiles_hifi = (strlen(request)>=12 && !strncmp(request,"PROFILES HI|",12));
#endif  // SUPPORT_BRISTOL
    int bdaddr_offset = 0;
    char* bdaddr = &request[9];  // bdaddr starts at pos 9 by default

#ifdef SUPPORT_BRISTOL    
    if (profiles_headset || profiles_protect || profiles_a2dp || profiles_hifi)
    {
      char* b = strstr(request, profiles_headset ? "HS|" : profiles_a2dp ? "A2DP|" : profiles_protect ? "HF|" : "HI|"); b+= (profiles_a2dp ? 5 : 3);
#else
    if (profiles_headset || profiles_protect || profiles_a2dp)
    {
      char* b = strstr(request, profiles_headset ? "HS|" : profiles_a2dp ? "A2DP|" : "HF|"); b+= (profiles_a2dp ? 5 : 3);
#endif  // SUPPORT_BRISTOL
      char* e = b;
      while (*e && *e!='|') e++;
      int len=e-b; if (len>63) len=63;
      char pin[64]; strncpy(pin, b, len); pin[len]=0;
      bdaddr = e+(*e=='|' ? 1 : 0);  // bdaddr is specified after the pin

      // enable pairing temporarily, some headsets/phones need a pin during profile discovery (might need to pair first)
      system("killall -SIGUSR2 hcid");
    
      // update client pincode file
      char buf[128];
      sprintf(buf, "echo %s > %s", pin, BT_CLIENT_PINFILE);
      system(buf);

      LS_DEBUG("blueserver: Setting [%s] in [%s]\n", pin, BT_CLIENT_PINFILE);
    }
        
    scan_profiles_fd = fd;
    LS_DEBUG("blueserver: Scanning [%s]\n",bdaddr);
#ifdef SUPPORT_BRISTOL
		if (profiles_hifi)
			scan_profiles_cmd_sp_browse(bdaddr);
		else
#endif  // SUPPORT_BRISTOL
			scan_profiles_cmd_browse(bdaddr);
    writestr(fd, "END");
    
    if (profiles_headset || profiles_protect || profiles_a2dp)
    {
      unlink(BT_CLIENT_PINFILE);  // remove client pinfile (so hcid uses default pin)
      system("killall -SIGUSR1 hcid");  // disable pairing again
    }
  }
  else if (!strncmp(request,"ENABLE",6))
  {
    EnableBluetooth(1);
    writestr(fd, "END");
  }
  else if (!strncmp(request,"DISABLE",7))
  {
    EnableBluetooth(0);
    writestr(fd, "END");
  }
  else if (!strncmp(request,"MSRS ENABLE",11))
  {
    LS_DEBUG("blueserver: MSRS ENABLE\n");

    system("hciconfig hci0 lm accept");
    system("hciconfig hci0 lp rswitch,hold,sniff,park");
    writestr(fd, "END");
  }
  else if (!strncmp(request,"MSRS DISABLE",12))
  {
    LS_DEBUG("blueserver: MSRS DISABLE\n");

    system("hciconfig hci0 lm master");
    system("hciconfig hci0 lp hold,sniff,park");
    writestr(fd, "END");
  }
  else if (!strncmp(request,"PAIRING ENABLE",14))
  {
    // "PAIRING ENABLE|<pin>"
    LS_DEBUG("blueserver: PAIRING ENABLE\n");

    // enable pairing temporarily, most devices need a pin during profile discovery or first connection
    system("killall -SIGUSR2 hcid");

    unlink(BT_CLIENT_PINFILE);
    if (strlen(request)>15 && !strncmp(request,"PAIRING ENABLE|",15))
    {
      char* b = strstr(request, "PAIRING ENABLE|"); b+= 15;
      char* e = b;
      while (*e && *e!='|') e++;
      int len=e-b; if (len>63) len=63;
      char pin[64]; strncpy(pin, b, len); pin[len]=0;

      // update client pincode file
      if (strlen(pin))
      {
        LS_DEBUG("blueserver: Writing PIN [%s] to %s\n", pin, BT_CLIENT_PINFILE);

        char buf[128];
        sprintf(buf, "echo %s > %s", pin, BT_CLIENT_PINFILE);
        system(buf);
      }
    }

    writestr(fd, "END");
  }
  else if (!strncmp(request,"PAIRING DISABLE",15))
  {
    LS_DEBUG("blueserver: PAIRING DISABLE\n");

    unlink(BT_CLIENT_PINFILE);  // remove client pinfile (so hcid uses default pin)
    system("killall -SIGUSR1 hcid");  // disable pairing again
  }
  else
  {
    writestr(fd, "END");
  }
}

int main(int argc,char** argv)
{
  if (argc<2) exit(-1);
  char* the_address = argv[1];
  return localserver(the_address,requesthandler);
}
